@@ -16,6 +16,7 @@
 #include <sys/stat.h>
 #include <sys/ioctl.h>
 #include <sys/utsname.h>
+#include <sys/wait.h>
 #include <arpa/inet.h>
 #include <net/if.h>
 #include <time.h>
@@ -23,10 +24,25 @@
 
 #define DHCP_OPTION_LEN 1236 /* pump 0.8.24 defines a max option size of 57,
                                 dhcp 2.0pl5 uses 1222, dhcp3 3.0.6 uses 1236 */
-#define DHCLIENT_REQUEST_DEFAULTS \
-  "subnet-mask, broadcast-address, time-offset, routers, domain-name, \
-   domain-name-servers, host-name"
-#define DHCLIENT_REQUEST_EXTRAS "ntp-servers"
+
+const char* dhclient_request_options_dhclient[] = { "subnet-mask",
+                                                    "broadcast-address",
+                                                    "time-offset",
+                                                    "routers",
+                                                    "domain-name",
+                                                    "domain-name-servers",
+                                                    "host-name",
+                                                    "ntp-servers", /* extra */
+                                                    NULL };
+
+const char* dhclient_request_options_udhcpc[] = { "subnet",
+                                                  "broadcast",
+                                                  "router",
+                                                  "domain",
+                                                  "hostname",
+                                                  "dns",
+                                                  "ntpsrv", /* extra */
+                                                  NULL };
 
 static int dhcp_exit_status = 1;
 static pid_t dhcp_pid = -1;
@@ -38,7 +54,7 @@ static pid_t dhcp_pid = -1;
 static void netcfg_write_dhcp (char *iface, char *dhostname)
 {
     FILE *fp;
-    
+
     if ((fp = file_open(INTERFACES_FILE, "a"))) {
         fprintf(fp, "\n# The primary network interface\n");
         if (!iface_is_hotpluggable(iface) && !find_in_stab(iface))
@@ -60,20 +76,21 @@ static void netcfg_write_dhcp (char *iface, char *dhostname)
         }
         fclose(fp);
     }
-    
-#if 0
-    if ((fp = file_open(RESOLV_FILE, "a"))) {
-        fclose(fp);
-    }
-#endif
 }
 
 /* Returns 1 if no default route is available */
 static short no_default_route (void)
 {
+#if defined(__FreeBSD_kernel__)
+    int status;
+
+    status = system("exec /lib/freebsd/route show default >/dev/null 2>&1");
+
+    return WEXITSTATUS(status) != 0;
+#else
     FILE* iproute = NULL;
     char buf[256] = { 0 };
-    
+
     if ((iproute = popen("ip route", "r")) != NULL) {
         while (fgets (buf, 256, iproute) != NULL) {
             if (buf[0] == 'd' && strstr (buf, "default via ")) {
@@ -83,8 +100,9 @@ static short no_default_route (void)
         }
         pclose(iproute);
     }
-    
+
     return 1;
+#endif
 }
 
 /*
@@ -94,7 +112,7 @@ static short no_default_route (void)
  * lease or because it succeeded and daemonized itself), this
  * gets the child's exit status and sets dhcp_pid to -1
  */
-static void dhcp_client_sigchld(int sig __attribute__ ((unused))) 
+static void dhcp_client_sigchld(int sig __attribute__ ((unused)))
 {
     if (dhcp_pid <= 0)
         return;
@@ -106,7 +124,7 @@ static void dhcp_client_sigchld(int sig __attribute__ ((unused)))
 }
 
 
-/* 
+/*
  * This function will start whichever DHCP client is available
  * using the provided DHCP hostname, if supplied
  *
@@ -115,24 +133,31 @@ static void dhcp_client_sigchld(int sig __attribute__ ((unused)))
 int start_dhcp_client (struct debconfclient *client, char* dhostname)
 {
     FILE *dc = NULL;
-    enum { DHCLIENT, DHCLIENT3, PUMP } dhcp_client;
-    
-    if (access("/var/lib/dhcp3", F_OK) == 0)
-        dhcp_client = DHCLIENT3;
-    else if (access("/sbin/dhclient", F_OK) == 0)
-        dhcp_client = DHCLIENT;
+    const char **ptr;
+    char **arguments;
+    int options_count;
+    enum { DHCLIENT, PUMP, UDHCPC } dhcp_client;
+    int dhcp_seconds;
+
+    if (access("/sbin/dhclient", F_OK) == 0)
+		dhcp_client = DHCLIENT;
     else if (access("/sbin/pump", F_OK) == 0)
         dhcp_client = PUMP;
+    else if (access("/sbin/udhcpc", F_OK) == 0)
+        dhcp_client = UDHCPC;
     else {
         debconf_input(client, "critical", "netcfg/no_dhcp_client");
         debconf_go(client);
         exit(1);
     }
-    
+
+    debconf_get(client, "netcfg/dhcp_timeout");
+    dhcp_seconds = atoi(client->value);
+
     if ((dhcp_pid = fork()) == 0) { /* child */
         /* disassociate from debconf */
         fclose(client->out);
-        
+
         /* get dhcp lease */
         switch (dhcp_client) {
         case PUMP:
@@ -140,44 +165,74 @@ int start_dhcp_client (struct debconfclient *client, char* dhostname)
                 execlp("pump", "pump", "-i", interface, "-h", dhostname, NULL);
             else
                 execlp("pump", "pump", "-i", interface, NULL);
-            
+
             break;
-            
+
         case DHCLIENT:
             /* First, set up dhclient.conf */
-            
             if ((dc = file_open(DHCLIENT_CONF, "w"))) {
-                fprintf(dc, "send dhcp-class-identifier \"d-i\";\n" );
-                fprintf(dc, "request " DHCLIENT_REQUEST_DEFAULTS", " \
-                                       DHCLIENT_REQUEST_EXTRAS";\n" );
-                if (dhostname) {
-                    fprintf(dc, "send host-name \"%s\";\n", dhostname);
-                }
-                fclose(dc);
-            }
-            
-            execlp("dhclient", "dhclient", "-e", interface, NULL);
-            break;
-            
-        case DHCLIENT3:
-            /* Different place.. */
-            
-            if ((dc = file_open(DHCLIENT3_CONF, "w"))) {
                 fprintf(dc, "send vendor-class-identifier \"d-i\";\n" );
-                fprintf(dc, "request " DHCLIENT_REQUEST_DEFAULTS", " \
-                                       DHCLIENT_REQUEST_EXTRAS";\n" );
+                fprintf(dc, "request ");
+
+                for (ptr = dhclient_request_options_dhclient; *ptr; ptr++) {
+                    fprintf(dc, *ptr);
+
+                    /* look ahead to see if it is the last entry */
+                    if (*(ptr + 1))
+                        fprintf(dc, ", ");
+                    else
+                        fprintf(dc, ";\n");
+                }
+
                 if (dhostname) {
                     fprintf(dc, "send host-name \"%s\";\n", dhostname);
                 }
+                fprintf(dc, "timeout %d;\n", dhcp_seconds);
                 fclose(dc);
             }
-            
+
             execlp("dhclient", "dhclient", "-1", interface, NULL);
             break;
+
+        case UDHCPC:
+            /* figure how many options we have */
+            options_count = 0;
+            for (ptr = dhclient_request_options_udhcpc; *ptr; ptr++)
+                options_count++;
+
+            /* Allow space for:
+                options: options_count * 2
+                 params: 5
+               hostname: 2
+                   NULL: 1 */
+            arguments = malloc((options_count * 2 + 5 + 2 + 1) * sizeof(char **));
+
+            /* set the command options */
+            options_count = 0;
+            arguments[options_count++] = "udhcpc";
+            arguments[options_count++] = "-i";
+            arguments[options_count++] = interface;
+            arguments[options_count++] = "-V";
+            arguments[options_count++] = "d-i";
+            for (ptr = dhclient_request_options_udhcpc; *ptr; ptr++) {
+                arguments[options_count++] = "-O";
+                arguments[options_count++] = *ptr;
+            }
+
+            if (dhostname) {
+                arguments[options_count++] = "-H";
+                arguments[options_count++] = dhostname;
+            }
+
+            arguments[options_count] = NULL;
+
+            execvp("udhcpc", arguments);
+            free(arguments);
+            break;
         }
         if (errno != 0)
             di_error("Could not exec dhcp client: %s", strerror(errno));
-        
+
         return 1; /* should NEVER EVER get here */
     }
     else if (dhcp_pid == -1)
@@ -192,7 +247,7 @@ int start_dhcp_client (struct debconfclient *client, char* dhostname)
 
 static int kill_dhcp_client(void)
 {
-    system("killall.sh"); 
+    system("killall.sh");
     return 0;
 }
 
@@ -213,18 +268,18 @@ int poll_dhcp_client (struct debconfclient *client)
     int seconds_slept = 0;
     int ret = 1;
     int dhcp_seconds;
-    
+
     debconf_get(client, "netcfg/dhcp_timeout");
-    
+
     dhcp_seconds = atoi(client->value);
-    
+
     /* show progress bar */
     debconf_capb(client, "backup progresscancel");
     debconf_progress_start(client, 0, dhcp_seconds, "netcfg/dhcp_progress");
     if (debconf_progress_info(client, "netcfg/dhcp_progress_note") == 30)
         goto stop;
     netcfg_progress_displayed = 1;
-    
+
     /* wait between 2 and dhcp_seconds seconds for a DHCP lease */
     while ( ((dhcp_pid > 0) || (seconds_slept < 2))
             && (seconds_slept < dhcp_seconds) ) {
@@ -234,11 +289,11 @@ int poll_dhcp_client (struct debconfclient *client)
             goto stop;
     }
     /* Either the client exited or time ran out */
-    
+
     /* got a lease? display a success message */
     if (!(dhcp_pid > 0) && (dhcp_exit_status == 0)) {
         ret = 0;
-        
+
         debconf_capb(client, "backup"); /* stop displaying cancel button */
         if (debconf_progress_set(client, dhcp_seconds) == 30)
             goto stop;
@@ -246,13 +301,13 @@ int poll_dhcp_client (struct debconfclient *client)
             goto stop;
         sleep(2);
     }
-    
+
  stop:
     /* stop progress bar */
     debconf_progress_stop(client);
     debconf_capb(client, "backup");
     netcfg_progress_displayed = 0;
-    
+
     return ret;
 }
 
@@ -263,27 +318,29 @@ int poll_dhcp_client (struct debconfclient *client)
 #define REPLY_DONT_CONFIGURE         3
 #define REPLY_RECONFIGURE_WIFI       4
 #define REPLY_LOOP_BACK              5
+#define REPLY_CHECK_DHCP             6
+#define REPLY_ASK_OPTIONS            7
 
 int ask_dhcp_options (struct debconfclient *client)
 {
     int ret;
-    
+
     if (is_wireless_iface(interface)) {
         debconf_metaget(client, "netcfg/internal-wifireconf", "description");
         debconf_subst(client, "netcfg/dhcp_options", "wifireconf", client->value);
     }
     else /* blank from last time */
         debconf_subst(client, "netcfg/dhcp_options", "wifireconf", "");
-    
+
     /* critical, we don't want to enter a loop */
     debconf_input(client, "critical", "netcfg/dhcp_options");
     ret = debconf_go(client);
-    
+
     if (ret == 30)
         return GO_BACK;
-    
+
     debconf_get(client, "netcfg/dhcp_options");
-    
+
     /* strcmp sucks */
     if (client->value[0] == 'R') {	/* _R_etry ... or _R_econfigure ... */
         size_t len = strlen(client->value);
@@ -302,16 +359,38 @@ int ask_dhcp_options (struct debconfclient *client)
         return REPLY_DONT_CONFIGURE;
 }
 
+int ask_wifi_configuration (struct debconfclient *client)
+{
+    enum { ABORT, DONE, ESSID, WEP } wifistate = ESSID;
+    for (;;) {
+        switch (wifistate) {
+        case ESSID:
+            wifistate = (netcfg_wireless_set_essid(client, interface, "high") == GO_BACK) ?
+                ABORT : WEP;
+            break;
+        case WEP:
+            wifistate = (netcfg_wireless_set_wep(client, interface) == GO_BACK) ?
+                ESSID : DONE;
+            break;
+        case ABORT:
+            return REPLY_ASK_OPTIONS;
+            break;
+        case DONE:
+            return REPLY_CHECK_DHCP;
+            break;
+        }
+    }
+}
+
 
-/* Here comes another Satan machine. */
 int netcfg_activate_dhcp (struct debconfclient *client)
 {
     char* dhostname = NULL;
     enum { START, POLL, ASK_OPTIONS, DHCP_HOSTNAME, HOSTNAME, DOMAIN, HOSTNAME_SANS_NETWORK } state = START;
-    
+
     kill_dhcp_client();
     loop_setup();
-    
+
     for (;;) {
         switch (state) {
         case START:
@@ -337,28 +416,31 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                  */
 
                 /* Before doing anything else, check for a default route */
-                
+
                 if (no_default_route()) {
                     debconf_input(client, "critical", "netcfg/no_default_route");
                     debconf_go(client);
                     debconf_get(client, "netcfg/no_default_route");
-                    
+
                     if (!strcmp(client->value, "false")) {
                         state = ASK_OPTIONS;
                         break;
                     }
                 }
-                
+
                 /*
                  * Set defaults for domain name and hostname
                  */
-                
+
+#ifndef MAXHOSTNAMELEN
+#define MAXHOSTNAMELEN 63
+#endif
                 char buf[MAXHOSTNAMELEN + 1] = { 0 };
                 char *ptr = NULL;
                 FILE *d = NULL;
-                
+
                 have_domain = 0;
-                
+
                 /*
                  * Default to the domain name returned via DHCP, if any
                  */
@@ -367,7 +449,7 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                     fgets(domain, _UTSNAME_LENGTH, d);
                     fclose(d);
                     unlink(DOMAIN_FILE);
-                    
+
                     if (!empty_str(domain) && verify_hostname(domain) == 0) {
                         debconf_set(client, "netcfg/get_domain", domain);
                         have_domain = 1;
@@ -383,9 +465,9 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                     fgets(ntpservers, DHCP_OPTION_LEN, d);
                     fclose(d);
                     unlink(NTP_SERVER_FILE);
-                    
+
                     if (!empty_str(ntpservers)) {
-                        debconf_set(client, "netcfg/dhcp_ntp_servers", 
+                        debconf_set(client, "netcfg/dhcp_ntp_servers",
                                     ntpservers);
                     }
                 }
@@ -409,7 +491,7 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                 } else {
                     struct ifreq ifr;
                     struct in_addr d_ipaddr = { 0 };
-                    
+
                     ifr.ifr_addr.sa_family = AF_INET;
                     strncpy(ifr.ifr_name, interface, IFNAMSIZ);
                     if (ioctl(skfd, SIOCGIFADDR, &ifr) == 0) {
@@ -419,7 +501,7 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                     else
                         di_warning("ioctl failed (%s)", strerror(errno));
                 }
-                
+
                 /*
                  * Default to the domain name that is the domain part
                  * of the hostname, if any
@@ -428,23 +510,24 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                     debconf_set(client, "netcfg/get_domain", ptr + 1);
                     have_domain = 1;
                 }
-                
+
                 /* Make sure we have NS going if the DHCP server didn't serve it up */
                 if (resolv_conf_entries() <= 0) {
                     char *nameservers = NULL;
-                    
+
                     if (netcfg_get_nameservers (client, &nameservers) == GO_BACK) {
                         state = ASK_OPTIONS;
                         break;
                     }
-                    
+
                     netcfg_nameservers_to_array (nameservers, nameserver_array);
+                    netcfg_write_resolv (domain, nameserver_array);
                 }
-                
+
                 state = HOSTNAME;
             }
             break;
-            
+
         case ASK_OPTIONS:
             /* DHCP client may still be running */
             switch (ask_dhcp_options (client)) {
@@ -472,35 +555,16 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                 }
                 break;
             case REPLY_RECONFIGURE_WIFI:
-                {
-                    /* oh god - a NESTED satan machine */
-                    enum { ABORT, DONE, ESSID, WEP } wifistate = ESSID;
-                    for (;;) {
-                        switch (wifistate) {
-                        case ESSID:
-                            wifistate = ( netcfg_wireless_set_essid(client, interface, "high") == GO_BACK ) ?
-                                ABORT : WEP;
-                            break;
-                        case WEP:
-                            wifistate = ( netcfg_wireless_set_wep (client, interface) == GO_BACK ) ?
-                                ESSID : DONE;
-                            break;
-                        case ABORT:
-                            state = ASK_OPTIONS;
-                            break;
-                        case DONE:
-                            if (dhcp_pid > 0)
-                                state = POLL;
-                            else {
-                                kill_dhcp_client();
-                                state = START;
-                            }
-                            break;
-                        }
-                        if (wifistate == DONE || wifistate == ABORT)
-                            break;
+                if (ask_wifi_configuration(client) == REPLY_CHECK_DHCP) {
+                    if (dhcp_pid > 0)
+                        state = POLL;
+                    else {
+                        kill_dhcp_client();
+                        state = START;
                     }
                 }
+                else
+                    state = ASK_OPTIONS;
                 break;
             }
             break;
@@ -518,7 +582,7 @@ int netcfg_activate_dhcp (struct debconfclient *client)
                 state = START;
             }
             break;
-            
+
         case HOSTNAME:
             if (netcfg_get_hostname (client, "netcfg/get_hostname", &hostname, 1)) {
                 /*
@@ -532,17 +596,27 @@ int netcfg_activate_dhcp (struct debconfclient *client)
             else
                 state = DOMAIN;
             break;
-            
+
         case DOMAIN:
             if (!have_domain && netcfg_get_domain (client, &domain))
                 state = HOSTNAME;
             else {
                 netcfg_write_common(ipaddress, hostname, domain);
                 netcfg_write_dhcp(interface, dhostname);
+                /* If the resolv.conf was written by udhcpc, then nameserver_array
+                 * will be empty and we'll need to populate it.  If we asked for
+                 * the nameservers, then it'll be full, but nobody will care if we
+                 * refill it.
+                 */
+                if (read_resolv_conf_nameservers(nameserver_array))
+                    netcfg_write_resolv(domain, nameserver_array);
+                else
+                    printf("Error reading resolv.conf for nameservers\n");
+
                 return 0;
             }
             break;
-            
+
         case HOSTNAME_SANS_NETWORK:
             if (netcfg_get_hostname (client, "netcfg/get_hostname", &hostname, 0))
                 state = ASK_OPTIONS;
@@ -555,31 +629,69 @@ int netcfg_activate_dhcp (struct debconfclient *client)
             break;
         }
     }
-} 
+}
 
 /* returns number of 'nameserver' entries in resolv.conf */
 int resolv_conf_entries (void)
 {
     FILE *f;
     int count = 0;
-    
+
     if ((f = fopen(RESOLV_FILE, "r")) != NULL) {
         char buf[256];
-        
+
         while (fgets(buf, 256, f) != NULL) {
             char *ptr;
-            
+
             if ((ptr = strchr(buf, ' ')) != NULL) {
                 *ptr = '\0';
                 if (strcmp(buf, "nameserver") == 0)
                     count++;
             }
         }
-        
+
         fclose(f);
     }
     else
         count = -1;
-    
+
     return count;
 }
+
+/* Read the nameserver entries out of resolv.conf and stick them into
+ * nameservers_array, so we can write out a newer, shinier resolv.conf
+ */
+int read_resolv_conf_nameservers(struct in_addr array[])
+{
+    FILE *f;
+    int i = 0;
+    
+    if ((f = fopen(RESOLV_FILE, "r")) != NULL) {
+        char buf[256];
+
+        while (fgets(buf, 256, f) != NULL) {
+            char *ptr;
+
+            if (strncmp(buf, "nameserver ", strlen("nameserver ")) == 0) {
+                /* Chop off trailing \n */
+                if (buf[strlen(buf)-1] == '\n')
+                    buf[strlen(buf)-1] = '\0';
+
+                ptr = buf + strlen("nameserver ");
+                inet_pton(AF_INET, ptr, &array[i++]);
+                if (i == 3) {
+                    /* We can only hold so many nameservers, and we've reached
+                     * our limit.  Sorry.
+                     */
+                    break;
+                }
+            }
+        }
+
+        fclose(f);
+        array[i].s_addr = 0;
+        return 1;
+    }
+    else
+        return 0;
+}
