@@ -1,20 +1,20 @@
-/* 
-   netcfg-common.c - Shared functions used to configure the network for 
+/*
+   netcfg-common.c - Shared functions used to configure the network for
    the debian-installer.
 
    Copyright (C) 2000-2002  David Kimdon <dwhedon@debian.org>
                             and others (see debian/copyright)
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
@@ -24,10 +24,9 @@
 #include "netcfg.h"
 #if defined(WIRELESS)
 #include <iwlib.h>
-#elif defined(__linux__)
-#include <linux/if.h>
 #endif
 #include <net/if_arp.h>
+#include <net/if.h>
 #include <errno.h>
 #include <assert.h>
 #include <ctype.h>
@@ -45,6 +44,14 @@
 #include <time.h>
 #include <netdb.h>
 
+#include <ifaddrs.h>
+
+#ifdef __FreeBSD_kernel__
+#define LO_IF	"lo0"
+#else
+#define LO_IF	"lo"
+#endif
+
 /* Set if there is currently a progress bar displayed. */
 int netcfg_progress_displayed = 0;
 
@@ -70,7 +77,7 @@ int inet_ptom (const char *src, int *dst, struct in_addr *addrp)
 {
     struct in_addr newaddr, *addr;
     in_addr_t mask, num;
-    
+
     if (src && !addrp) {
         if (inet_pton (AF_INET, src, &newaddr) < 0)
             return 0;
@@ -78,22 +85,22 @@ int inet_ptom (const char *src, int *dst, struct in_addr *addrp)
     }
     else
         addr = addrp;
-    
+
     mask = ntohl(addr->s_addr);
 
     for (num = mask; num & 1; num >>= 1);
-    
+
     if (num != 0 && mask != 0) {
         for (num = ~mask; num & 1; num >>= 1);
         if (num)
             return 0;
     }
-    
+
     for (num = 0; mask; mask <<= 1)
         num++;
-    
+
     *dst = num;
-    
+
     return 1;
 }
 
@@ -102,12 +109,12 @@ const char *inet_mtop (int src, char *dst, socklen_t cnt)
 {
     struct in_addr addr;
     in_addr_t mask = 0;
-    
+
     for(; src; src--)
         mask |= 1 << (32 - src);
-    
+
     addr.s_addr = htonl(mask);
-    
+
     return inet_ntop (AF_INET, &addr, dst, cnt);
 }
 
@@ -119,6 +126,8 @@ void open_sockets (void)
     skfd = socket (AF_INET, SOCK_DGRAM, 0);
 }
 
+#ifdef __linux__
+
 #define SYSCLASSNET "/sys/class/net/"
 
 /* Returns non-zero if this interface has an enabled kill switch, otherwise
@@ -133,24 +142,24 @@ int check_kill_switch(const char *iface)
     int linklen, killlen;
     int fd = -1;
     int ret = 0;
-    
+
     /* longest string we need */
     len = strlen(SYSCLASSNET) + strlen(iface) + strlen("/device/rf_kill") + 1;
-    
+
     temp = malloc(len);
     snprintf(temp, len, SYSCLASSNET "%s/driver", iface);
     linkbuf = malloc(1024); /* probably OK ... I hate readlink() */
     linklen = readlink(temp, linkbuf, 1024);
     if (linklen < 0)
         goto out;
-    
+
     if (strncmp(linkbuf + linklen - 8, "/ipw2100", 8) == 0)
         killname = "rf_kill";
     else if (strncmp(linkbuf + linklen - 8, "/ipw2200", 8) == 0)
         killname = "rf_kill";
     else
         goto out;
-    
+
     snprintf(temp, len, SYSCLASSNET "%s/device/%s", iface, killname);
     di_info("Checking RF kill switch: %s", temp);
     fd = open(temp, O_RDONLY);
@@ -164,12 +173,12 @@ int check_kill_switch(const char *iface)
         di_warning("RF kill state file empty");
         goto out;
     }
-    
+
     if (killstate == '2') {
         di_info("RF kill switch enabled");
         ret = 1;
     }
-    
+
  out:
     free(temp);
     free(linkbuf);
@@ -180,13 +189,21 @@ int check_kill_switch(const char *iface)
 
 #undef SYSCLASSNET
 
+#else /* !__linux__ */
+int check_kill_switch(const char *iface)
+{
+    return 0;
+}
+#endif /* __linux__ */
+
+#if defined(WIRELESS)
 int is_raw_80211(const char *iface)
 {
     struct ifreq ifr;
     struct sockaddr sa;
-    
+
     strncpy(ifr.ifr_name, iface, IFNAMSIZ);
-    
+
     if (skfd && ioctl(skfd, SIOCGIFHWADDR, &ifr) < 0) {
         di_warning("Unable to retrieve interface type.");
         return 0;
@@ -203,99 +220,125 @@ int is_raw_80211(const char *iface)
         return 0;
     }
 }
+#endif
 
-int is_interface_up(char *inter)
+int qsort_strcmp(const void *a, const void *b)
 {
-    struct ifreq ifr;
-    
-    strncpy(ifr.ifr_name, inter, sizeof(ifr.ifr_name));
-    
-    if (ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0)
-        return -1;
-    
-    return ((ifr.ifr_flags & IFF_UP) ? 1 : 0);
-}
-
-void get_name(char *name, char *p)
-{
-    while (isspace(*p))
-        p++;
-    while (*p) {
-        if (isspace(*p))
-            break;
-        if (*p == ':') {	/* could be an alias */
-            char *dot = p, *dotname = name;
-            *name++ = *p++;
-            while (isdigit(*p))
-                *name++ = *p++;
-            if (*p != ':') {	/* it wasn't, backup */
-                p = dot;
-                name = dotname;
-            }
-            if (*p == '\0')
-                return;
-            p++;
-            break;
-        }
-        *name++ = *p++;
-    }
-    *name++ = '\0';
-    return;
+    const char **ia = (const char **)a;
+    const char **ib = (const char **)b;
+    return strcmp(*ia, *ib);
 }
 
+#ifdef __GNU__
+#include <mach.h>
+#include <device/device.h>
+#include <hurd.h>
+/* On Hurd, the IP stack (pfinet) does not know the list of network interfaces
+ * before we configure them, so we cannot use getifaddrs(). Instead we try
+ * possible names for network interfaces and check whether they exists by
+ * attempting to open the kernel device. */
 int get_all_ifs (int all, char*** ptr)
 {
-    FILE *ifs = NULL;
-    char ibuf[512], rbuf[512];
+    static const char *const fmt[] = { "eth%d", "wl%d", NULL };
+
+    mach_port_t device_master;
+    device_t device;
+    int err;
+    char **list;
+    int num, i, j;
+    char name[3 + 3 * sizeof (int) + 1];
+
+    err = get_privileged_ports (0, &device_master);
+    if (err)
+	return 0;
+
+    num = 0;
+    list = malloc(sizeof *list);
+    for (i = 0; fmt[i]; i++)
+	for (j = 0;; j++) {
+	    sprintf (name, fmt[i], j);
+	    err = device_open (device_master, D_READ, name, &device);
+	    if (err != 0)
+		break;
+
+	    device_close (device);
+	    mach_port_deallocate (mach_task_self (), device);
+
+	    list = realloc (list, (num + 2) * sizeof *list);
+	    list[num++] = strdup(name);
+	}
+    list[num] = NULL;
+
+    mach_port_deallocate (mach_task_self (), device_master);
+    *ptr = list;
+    return num;
+}
+#else
+int get_all_ifs (int all, char*** ptr)
+{
+    struct ifaddrs *ifap, *ifa;
+    char ibuf[512];
     char** list = NULL;
     size_t len = 0;
-    
-    if ((ifs = fopen("/proc/net/dev", "r")) != NULL) {
-        fgets(ibuf, sizeof(ibuf), ifs); /* eat header */
-        fgets(ibuf, sizeof(ibuf), ifs); /* ditto */
-    }
-    else
+
+    if (getifaddrs(&ifap) == -1)
         return 0;
 
-    while (fgets(rbuf, sizeof(rbuf), ifs) != NULL) {
-        get_name(ibuf, rbuf);
-        if (!strcmp(ibuf, "lo"))        /* ignore the loopback */
+    for (ifa = ifap; ifa; ifa = ifa->ifa_next) {
+        strncpy(ibuf, ifa->ifa_name, sizeof(ibuf));
+        if (ifa->ifa_flags & IFF_LOOPBACK)   /* ignore loopback devices */
             continue;
+#if defined(__linux__)
         if (!strncmp(ibuf, "sit", 3))        /* ignore tunnel devices */
             continue;
+#endif
+#if defined(WIRELESS)
         if (is_raw_80211(ibuf))
             continue;
-        if (all || is_interface_up(ibuf) == 1) {
-            list = realloc(list, sizeof(char*) * (len + 1));
-            list[len] = strdup(ibuf);
-            len++;
+#endif
+        if (all || ifa->ifa_flags & IFF_UP) {
+            int found = 0;
+            size_t i;
+
+            for (i = 0 ; i < len ; i++) {
+                if (!strcmp(ibuf, list[i])) {
+                    found = 1;
+                }
+            }
+            if (!found) {
+                list = realloc(list, sizeof(char*) * (len + 2));
+                list[len] = strdup(ibuf);
+                len++;
+            }
         }
     }
-    
-    /* OK, now terminate it if necessary */
+
+    /* OK, now sort the list and terminate it if necessary */
     if (list != NULL) {
-        list = realloc(list, sizeof(char*) * (len + 1));
+        qsort(list, len, sizeof(char *), qsort_strcmp);
         list[len] = NULL;
     }
-    fclose (ifs);
-    
+    freeifaddrs(ifap);
+
     *ptr = list;
-    
+
     return len;
 }
+#endif
 
+#ifdef __linux__
 short find_in_stab(const char* iface)
 {
     FILE *dn = NULL;
     char buf[128];
     size_t len = strlen(iface);
-    
+
     if (access(STAB, F_OK) == -1)
         return 0;
-    
+
     if (!(dn = popen("grep -v '^Socket' " STAB " | cut -f5", "r")))
         return 0;
-    
+
     while (fgets (buf, 128, dn) != NULL) {
         if (!strncmp(buf, iface, len)) {
             pclose(dn);
@@ -305,19 +348,26 @@ short find_in_stab(const char* iface)
     pclose(dn);
     return 0;
 }
+#else /* !__linux__ */
+/* Stub function for platforms not supporting /var/run/stab. */
+short find_in_stab(const char* iface)
+{
+    return 0;
+}
+#endif /* __linux__ */
 
 char *find_in_devnames(const char* iface)
 {
     FILE* dn = NULL;
     char buf[512], *result = NULL;
     size_t len = strlen(iface);
-    
+
     if (!(dn = fopen(DEVNAMES, "r")))
         return NULL;
-    
+
     while (fgets(buf, 512, dn) != NULL) {
         char *ptr = strchr(buf, ':'), *desc = ptr + 1;
-        
+
         if (!ptr) {
             result = NULL; /* corrupt */
             break;
@@ -327,35 +377,35 @@ char *find_in_devnames(const char* iface)
             break;
         }
     }
-    
+
     fclose(dn);
-    
+
     if (result) {
         len = strlen(result);
-        
+
         if (result[len - 1] == '\n')
             result[len - 1] = '\0';
     }
-    
+
     return result;
 }
 
 char *get_ifdsc(struct debconfclient *client, const char *ifp)
 {
     char template[256], *ptr = NULL;
-    
+
     if ((ptr = find_in_devnames(ifp)) != NULL) {
         debconf_metaget(client, "netcfg/internal-wireless", "description");
-        
+
         if (is_wireless_iface(ifp)) {
             size_t len = strlen(ptr) + strlen(client->value) + 4;
             ptr = realloc(ptr, len);
-            
+
             di_snprintfcat(ptr, len, " (%s)", client->value);
         }
         return ptr; /* already strdup'd */
     }
-    
+
     if (strlen(ifp) < 100) {
         if (!is_wireless_iface(ifp)) {
             /* strip away the number from the interface (eth0 -> eth) */
@@ -363,10 +413,10 @@ char *get_ifdsc(struct debconfclient *client, const char *ifp)
             while ((*ptr < '0' || *ptr > '9') && *ptr != '\0')
                 ptr++;
             *ptr = '\0';
-            
+
             sprintf(template, "netcfg/internal-%s", new_ifp);
             free(new_ifp);
-            
+
             if (debconf_metaget(client, template, "description") == 0 &&
                 client->value != NULL) {
                 return strdup(client->value);
@@ -389,12 +439,12 @@ int iface_is_hotpluggable(const char *iface)
     FILE* f = NULL;
     char buf[256];
     size_t len = strlen(iface);
-    
+
     if (!(f = fopen(DEVHOTPLUG, "r"))) {
         di_info("No hotpluggable devices are present in the system.");
         return 0;
     }
-    
+
     while (fgets(buf, 256, f) != NULL) {
         if (!strncmp(buf, iface, len)) {
             di_info("Detected %s as a hotpluggable device", iface);
@@ -402,9 +452,9 @@ int iface_is_hotpluggable(const char *iface)
             return 1;
         }
     }
-    
+
     fclose(f);
-    
+
     di_info("Hotpluggable devices available, but %s is not one of them", iface);
     return 0;
 }
@@ -433,7 +483,7 @@ void netcfg_die(struct debconfclient *client)
 
 /**
  * @brief Ask which interface to configure
- * @param client - client 
+ * @param client - client
  * @param interface      - set to the answer
  * @param numif - number of interfaces found.
  */
@@ -448,35 +498,35 @@ int netcfg_get_interface(struct debconfclient *client, char **interface,
     char *ptr = NULL;
     char *ifdsc = NULL;
     char *old_selection = NULL;
-    
+
     if (*interface) {
         free(*interface);
         *interface = NULL;
     }
-    
+
     if (!(ptr = malloc(128)))
         goto error;
-    
+
     len = 128;
     *ptr = '\0';
-    
+
     num_interfaces = get_all_ifs(1, &ifs);
-    
+
     /* If no default was provided, use the first in the list of interfaces. */
     if (! defif && num_interfaces > 0) {
         defif=ifs[0];
     }
-    
+
     /* Remember old interface selection, in case it's preseeded. */
     debconf_get(client, "netcfg/choose_interface");
     old_selection = strdup(client->value);
-    
+
     for (i = 0; i < num_interfaces; i++) {
         size_t newchars;
         char *temp = NULL;
-        
+
         inter = ifs[i];
-        
+
         interface_down(inter);
         ifdsc = get_ifdsc(client, inter);
         newchars = strlen(inter) + strlen(ifdsc) + 5; /* ": , " + NUL */
@@ -485,21 +535,21 @@ int netcfg_get_interface(struct debconfclient *client, char **interface,
                 goto error;
             len += newchars + 128;
         }
-        
+
         temp = malloc(newchars);
-        
+
         snprintf(temp, newchars, "%s: %s", inter, ifdsc);
-        
+
         if (num_interfaces > 1 &&
             ((strcmp(defif, inter) == 0) || (strcmp(defif, temp) == 0)))
             debconf_set(client, "netcfg/choose_interface", temp);
-        
+
         di_snprintfcat(ptr, len, "%s, ", temp);
-        
+
         free(temp);
         free(ifdsc);
     }
-    
+
     if (num_interfaces == 0) {
         debconf_input(client, "high", "netcfg/no_interfaces");
         ret = debconf_go(client);
@@ -517,32 +567,32 @@ int netcfg_get_interface(struct debconfclient *client, char **interface,
         *numif = num_interfaces;
         /* remove the trailing ", ", which confuses cdebconf */
         ptr[strlen(ptr) - 2] = '\0';
-        
+
         debconf_subst(client, "netcfg/choose_interface", "ifchoices", ptr);
         free(ptr);
-        
+
         asked = (debconf_input(client, "critical", "netcfg/choose_interface") == 0);
         ret = debconf_go(client);
-        
+
         /* If the question is not asked, honor preseeded interface name.
          * However, if it was preseeded to "auto", or there was no old value,
          * leave it set to defif. */
         if (!asked && strlen(old_selection) && strcmp(old_selection, "auto") != 0) {
             debconf_set(client, "netcfg/choose_interface", old_selection);
         }
-        
+
         free(old_selection);
-        
+
         if (ret)
             return ret;
-        
+
         debconf_get(client, "netcfg/choose_interface");
         inter = client->value;
-        
+
         if (!inter)
             netcfg_die(client);
     }
-    
+
     /* grab just the interface name, not the description too */
     *interface = inter;
     /* Note that the question may be preseeded to just the interface name,
@@ -552,17 +602,17 @@ int netcfg_get_interface(struct debconfclient *client, char **interface,
         *ptr = '\0';
     }
     *interface = strdup(*interface);
-    
+
     /* Free allocated memory */
     while (ifs && *ifs)
         free(*ifs++);
-    
+
     return 0;
 
  error:
     if (ptr)
         free(ptr);
-    
+
     netcfg_die(client);
     return 10; /* unreachable */
 }
@@ -577,9 +627,9 @@ short verify_hostname (char *hname)
         "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-.";
     size_t len;
     assert(hname != NULL);
-    
+
     len = strlen(hname);
-    
+
     /* Check the hostname for RFC 1123 compliance.  */
     if ((len < 1) ||
         (len > 63) ||
@@ -593,25 +643,25 @@ short verify_hostname (char *hname)
 }
 
 /*
- * Set the hostname. 
+ * Set the hostname.
  * @return 0 on success, 30 on BACKUP being selected.
  */
 int netcfg_get_hostname(struct debconfclient *client, char *template, char **hostname, short hdset)
 {
     int ret;
     char *s;
-    
+
     for(;;) {
         if (hdset)
             have_domain = 0;
         debconf_input(client, "high", template);
         ret = debconf_go(client);
-        
+
         if (ret == 30) /* backup */
             return ret;
-        
+
         debconf_get(client, template);
-        
+
         if (verify_hostname(client->value) != 0) {
             debconf_subst(client, "netcfg/invalid_hostname",
                           "hostname", client->value);
@@ -626,7 +676,7 @@ int netcfg_get_hostname(struct debconfclient *client, char *template, char **hos
             break;
         }
     }
-    
+
     /* don't strip DHCP hostnames */
     if (hdset && (s = strchr(*hostname, '.'))) {
         if (s[1] == '\0') { /* "somehostname." <- . should be ignored */
@@ -637,6 +687,7 @@ int netcfg_get_hostname(struct debconfclient *client, char *template, char **hos
             domain = strdup(s + 1);
             debconf_set(client, "netcfg/get_domain", domain);
             have_domain = 1;
+            *s = '\0';
         }
     }
     return 0;
@@ -648,7 +699,7 @@ int netcfg_get_hostname(struct debconfclient *client, char *template, char **hos
 int netcfg_get_domain(struct debconfclient *client,  char **domain)
 {
     int ret;
-    
+
     if (have_domain == 1)
     {
         debconf_get(client, "netcfg/get_domain");
@@ -661,29 +712,33 @@ int netcfg_get_domain(struct debconfclient *client,  char **domain)
 
     debconf_input (client, "high", "netcfg/get_domain");
     ret = debconf_go(client);
-    
+
     if (ret)
         return ret;
-    
+
     debconf_get (client, "netcfg/get_domain");
-    
+
     if (*domain)
         free(*domain);
     *domain = NULL;
-    if (!empty_str(client->value))
-        *domain = strdup(client->value);
+    if (!empty_str(client->value)) {
+        const char *start = client->value;
+        while (*start == '.')
+            ++start; /* trim leading dots */
+        *domain = strdup(start);
+    }
     return 0;
 }
 
 void netcfg_write_loopback (void)
 {
     FILE *fp;
-    
+
     if ((fp = file_open(INTERFACES_FILE, "w"))) {
         fprintf(fp, HELPFUL_COMMENT);
         fprintf(fp, "\n# The loopback network interface\n");
-        fprintf(fp, "auto lo\n");
-        fprintf(fp, "iface lo inet loopback\n");
+        fprintf(fp, "auto "LO_IF"\n");
+        fprintf(fp, "iface "LO_IF" inet loopback\n");
         fclose(fp);
     }
 }
@@ -697,31 +752,31 @@ void netcfg_write_loopback (void)
 void netcfg_write_common(struct in_addr ipaddress, char *hostname, char *domain)
 {
     FILE *fp;
-    
+
     if (!hostname)
         return;
-    
+
     if ((fp = file_open(INTERFACES_FILE, "w"))) {
         fprintf(fp, HELPFUL_COMMENT);
         fprintf(fp, "\n# The loopback network interface\n");
-        fprintf(fp, "auto lo\n");
-        fprintf(fp, "iface lo inet loopback\n");
+        fprintf(fp, "auto "LO_IF"\n");
+        fprintf(fp, "iface "LO_IF" inet loopback\n");
         fclose(fp);
     }
-    
+
     /* Currently busybox, hostname is not available. */
     sethostname (hostname, strlen(hostname) + 1);
-    
+
     if ((fp = file_open(HOSTNAME_FILE, "w"))) {
         fprintf(fp, "%s\n", hostname);
         fclose(fp);
     }
-    
+
     if ((fp = file_open(HOSTS_FILE, "w"))) {
         char ptr1[INET_ADDRSTRLEN];
-        
+
         fprintf(fp, "127.0.0.1\tlocalhost\n");
-        
+
         if (ipaddress.s_addr) {
             inet_ntop (AF_INET, &ipaddress, ptr1, sizeof(ptr1));
             if (domain && !empty_str(domain))
@@ -734,9 +789,9 @@ void netcfg_write_common(struct in_addr ipaddress, char *hostname, char *domain)
             else
                 fprintf(fp, "127.0.1.1\t%s\n", hostname);
         }
-        
+
         fprintf(fp, "\n" IPV6_HOSTS);
-        
+
         fclose(fp);
     }
 }
@@ -745,22 +800,28 @@ void netcfg_write_common(struct in_addr ipaddress, char *hostname, char *domain)
 void deconfigure_network(void)
 {
     /* deconfiguring network interfaces */
-    interface_down("lo");
+    interface_down(LO_IF);
     interface_down(interface);
 }
 
 void loop_setup(void)
 {
     static int afpacket_notloaded = 1;
-    
+
     deconfigure_network();
-    
+
+#if defined(__FreeBSD_kernel__)
+    /* GNU/kFreeBSD currently uses the ifconfig command */
+    di_exec_shell_log("ifconfig "LO_IF" up");
+    di_exec_shell_log("ifconfig "LO_IF" 127.0.0.1 netmask 255.0.0.0");
+#else
     if (afpacket_notloaded)
         afpacket_notloaded = di_exec_shell("modprobe af_packet"); /* should become 0 */
-    
-    di_exec_shell_log("ip link set lo up");
-    di_exec_shell_log("ip addr flush dev lo");
-    di_exec_shell_log("ip addr add 127.0.0.1/8 dev lo");
+
+    di_exec_shell_log("ip link set "LO_IF" up");
+    di_exec_shell_log("ip addr flush dev "LO_IF);
+    di_exec_shell_log("ip addr add 127.0.0.1/8 dev "LO_IF);
+#endif
 }
 
 void seed_hostname_from_dns (struct debconfclient * client, struct in_addr *ipaddr)
@@ -768,43 +829,43 @@ void seed_hostname_from_dns (struct debconfclient * client, struct in_addr *ipad
     struct sockaddr_in sin;
     char *host;
     int err;
-    
+
     host = malloc(NI_MAXHOST);
     if (!host)
         netcfg_die(client);
-    
+
     /* copy IP address into required format */
     sin.sin_family = AF_INET;
     sin.sin_port = 0;
     memcpy(&sin.sin_addr, ipaddr, sizeof(*ipaddr));
-    
+
     /* attempt resolution */
     err = getnameinfo((struct sockaddr *) &sin, sizeof(sin),
                       host, NI_MAXHOST, NULL, 0, NI_NAMEREQD);
-    
+
     /* got it? */
     if (err == 0 && !empty_str(host)) {
         /* remove domain part */
         char* ptr = strchr(host, '.');
-        
+
         if (ptr)
             *ptr = '\0';
-        
+
         debconf_set(client, "netcfg/get_hostname", host);
-        
+
         if (!have_domain && (ptr && ptr[1] != '\0'))
             debconf_set(client, "netcfg/get_domain", ptr + 1);
     }
-    
+
     free(host);
 }
 
 void interface_up (char* iface)
 {
     struct ifreq ifr;
-    
+
     strncpy(ifr.ifr_name, iface, IFNAMSIZ);
-    
+
     if (skfd && ioctl(skfd, SIOCGIFFLAGS, &ifr) >= 0) {
         strncpy(ifr.ifr_name, iface, IFNAMSIZ);
         ifr.ifr_flags |= (IFF_UP | IFF_RUNNING);
@@ -815,9 +876,9 @@ void interface_up (char* iface)
 void interface_down (char* iface)
 {
     struct ifreq ifr;
-    
+
     strncpy(ifr.ifr_name, iface, IFNAMSIZ);
-    
+
     if (skfd && ioctl(skfd, SIOCGIFFLAGS, &ifr) >= 0) {
         strncpy(ifr.ifr_name, iface, IFNAMSIZ);
         ifr.ifr_flags &= ~IFF_UP;
@@ -835,7 +896,12 @@ void parse_args (int argc, char ** argv)
                 exit(EXIT_SUCCESS);
             }
         }
-        
+
+        if (!strcmp(argv[1], "write_loopback")) {
+            netcfg_write_loopback();
+            exit(EXIT_SUCCESS);
+        }
+
         exit(EXIT_FAILURE);
     }
 }
@@ -844,9 +910,9 @@ void reap_old_files (void)
 {
     static char* remove[] =
         { INTERFACES_FILE, HOSTS_FILE, HOSTNAME_FILE, NETWORKS_FILE,
-          RESOLV_FILE, DHCLIENT_CONF, DHCLIENT3_CONF, DOMAIN_FILE, 0 };
+          RESOLV_FILE, DHCLIENT_CONF, DOMAIN_FILE, 0 };
     char **ptr = remove;
-    
+
     while (*ptr)
         unlink(*ptr++);
 }
@@ -855,10 +921,10 @@ void netcfg_nameservers_to_array(char *nameservers, struct in_addr array[])
 {
     char *save, *ptr, *ns;
     int i;
-    
+
     if (nameservers) {
         save = ptr = strdup(nameservers);
-        
+
         for (i = 0; i < 3; i++) {
             ns = strtok_r(ptr, " \n\t", &ptr);
             if (ns)
@@ -866,7 +932,7 @@ void netcfg_nameservers_to_array(char *nameservers, struct in_addr array[])
             else
                 array[i].s_addr = 0;
         }
-        
+
         array[3].s_addr = 0;
         free(save);
     } else
@@ -877,7 +943,7 @@ int netcfg_get_nameservers (struct debconfclient *client, char **nameservers)
 {
     char *ptr, ptr1[INET_ADDRSTRLEN];
     int ret;
-    
+
     debconf_get(client,"netcfg/get_nameservers");
     if (*nameservers)
         ptr = *nameservers;
@@ -890,16 +956,16 @@ int netcfg_get_nameservers (struct debconfclient *client, char **nameservers)
     else
         ptr = "";
     debconf_set(client, "netcfg/get_nameservers", ptr);
-    
+
     debconf_input(client, "critical", "netcfg/get_nameservers");
     ret = debconf_go(client);
-    
+
     if (ret)
         return ret;
-    
+
     debconf_get(client, "netcfg/get_nameservers");
     ptr = client->value;
-    
+
     if (*nameservers)
         free(*nameservers);
     *nameservers = NULL;
@@ -907,3 +973,10 @@ int netcfg_get_nameservers (struct debconfclient *client, char **nameservers)
         *nameservers = strdup(ptr);
     return ret;
 }
+
+void netcfg_update_entropy (void)
+{
+#ifdef __linux__
+    di_exec_shell("ip addr show >/dev/random");
+#endif
+}
